.text
.arm

AGB_IRQVECT = 0x3007FFC

REG_BASE = 0x04000000
REG_DISPCNT = 0x00
REG_DISPSTAT = 0x04
REG_BG0CNT = 0x08
REG_BG1CNT = 0x0a
REG_BG2CNT = 0x0c
REG_BG0VOFS = 0x12
REG_BG1VOFS = 0x16
REG_BLDCNT = 0x50
REG_BLDALPHA = 0x52
REG_BLDY = 0x54
REG_P1 = 0x130

REG_IE = 0x200
REG_IF = 0x202
REG_IME = 0x208

MODE0 = 0x0
FORCE_BLANK = 0x80
BG0_EN = 0x100
BG1_EN = 0x200
BG2_EN = 0x400

COLOR16 = 0x0000
COLOR256 = 0x0080
SIZE256x256 = 0x0000
SIZE512x256 = 0x4000
SIZE256x512 = 0x8000
SIZE512x512 = 0xC000

A_BTN = 1
B_BTN = 2
SELECT = 4
START = 8
RIGHT = 16
LEFT = 32
UP = 64
DOWN = 128
R_BTN = 256
L_BTN = 512

MEM_PALETTE = 0x05000000
MEM_VRAM = 0x06000000
FONT_FIRSTCHAR = 32
FONT_CHAR_PAGE = 0
FONT_PALETTE_NUMBER = 14
BACKGROUND_CHAR_PAGE = 1
BACKGROUND_PALETTE_NUMBER = 0
BACKGROUND_TILEMAP_NUMBER = 3
UI_TILEMAP_NUMBER = 2
THROBBER_TILEMAP_NUMBER = 31
SCREENBASE = MEM_VRAM + UI_TILEMAP_NUMBER*2048
THROBBERBASE = MEM_VRAM + THROBBER_TILEMAP_NUMBER*2048
FONT_MEM = MEM_VRAM + FONT_CHAR_PAGE*16384 + FONT_FIRSTCHAR*32
FONT_PAL = MEM_PALETTE + FONT_PALETTE_NUMBER*16*2

BACKGROUND_MEM = MEM_VRAM + BACKGROUND_CHAR_PAGE*16384
BACKGROUND_PAL = MEM_PALETTE + BACKGROUND_PALETTE_NUMBER*16*2
BACKGROUND_TILE = MEM_VRAM + BACKGROUND_TILEMAP_NUMBER*2048

ARGUMENTS_OFFSET = 0x0203fbf8

MAGIC_KEY = 0xEBAB0BAF
MAGIC_KEY_OFFSET = 0x0203fff4

start:
.word variables

mov r0,#0xfc
swi 0x010000

@check for magic key, maybe
adr r1,argument_count
ldr r1,[r1]
cmp r1,#1
bgt normal_startup
adr r1,reset_code
ldr r1,[r1]
cmp r1,#0
beq normal_startup

ldr r2,=MAGIC_KEY
ldr r4,=MAGIC_KEY_OFFSET
ldr r3,[r4],#4
cmp r2,r3
bne normal_startup
ldr r3,[r4],#4
ldr r5,[r4],#4
eor r5,r5,r2
cmp r3,r5
bne normal_startup

@magic key found, so jump to reset code
bx r1

normal_startup:

@clear ram
mov r0,#0
mov r1,#0x02000000
ldr r2,=(1<<24|0xfff8)
swi 0x0C0000 @CpuFastSet

@copy to ewram if multiboot
adr r4,return_point
ldr r0,[r4]
tst r0,#0x02000000
beq non_multiboot

mov r0,#0x08000000
add r0,r0,#0xc0
ldr r2,[r0],#4
mov r1,#0x02000000
cmp r2,#1
beq multiboot_lz77
cmp r2,#2
beq multiboot_apack

adr r2,plugin_cpufastset
ldr r2,[r2]
swi 0x0C0000 @CpuFastSet
b non_multiboot

multiboot_lz77:
swi 0x110000 @L77UnCompWram
b non_multiboot

multiboot_apack:
bl depack

non_multiboot:
adr r6,argument_count
ldr r6,[r6]
lsl r6,r6,#2
add r6,r6,r6,LSL#1
cmp r6,#0
@skip menu on no arguments
beq return_to_plugin
mov r0,#0
adr r7,arguments_array
ldr r7,[r7]
cmp r6,#(12)
beq load_details

@set display
ldr r5,=REG_BASE
ldr r0,=(MODE0|BG0_EN|BG1_EN|BG2_EN|FORCE_BLANK)
strh r0,[r5,#REG_DISPCNT]
ldr r0,=(0|(FONT_CHAR_PAGE<<2)|(0<<6)|COLOR16|(UI_TILEMAP_NUMBER<<8)|SIZE256x256)
strh r0,[r5,#REG_BG0CNT]
ldr r0,=(1|(FONT_CHAR_PAGE<<2)|(0<<6)|COLOR16|(THROBBER_TILEMAP_NUMBER<<8)|SIZE256x256)
strh r0,[r5,#REG_BG1CNT]
ldr r0,=(2|(BACKGROUND_CHAR_PAGE<<2)|(0<<6)|COLOR256|(BACKGROUND_TILEMAP_NUMBER<<8)|SIZE256x256)
strh r0,[r5,#REG_BG2CNT]

load_background:
adr r4,background_char
ldr r8,[r4],#4
ldr r9,[r4],#4
ldr r10,[r4],#4
orr r11,r8,r9
orrs r11,r11,r10
beq load_font
mov r0,r8
ldr r1,=BACKGROUND_MEM
swi 0x120000 @L77UnCompVram
mov r0,r9
ldr r1,=BACKGROUND_PAL
swi 0x120000 @L77UnCompVram
mov r0,r10
ldr r1,=BACKGROUND_TILE
swi 0x120000 @L77UnCompVram

load_font:
adr r0,font
ldr r0,[r0]
ldr r1,=FONT_MEM
swi 0x120000 @L77UnCompVram
adr r0,fontpal
ldr r0,[r0]
ldr r1,=FONT_PAL
swi 0x120000 @L77UnCompVram

ldr r5,=REG_BASE
ldr r0,=(MODE0|BG0_EN|BG1_EN|BG2_EN)
strh r0,[r5,#REG_DISPCNT]

ldr r0,=((1<<(0+1))|(1<<6)|(1<<(8+2)))
strh r0,[r5,#REG_BLDCNT]
mov r0,#0
@ldr r0,=((8<<0)|(8<<8))
strh r0,[r5,#REG_BLDALPHA]

str r0,[r5,#REG_IME]

mrs r0,cpsr
bic r0,r0,#0x80  @clear "interrupt disable" bit
msr cpsr,r0

mov r0,#0x0008
strh r0,[r5,#REG_DISPSTAT]

ldr r0,=0xFFFF0001
str r0,[r5,#REG_IE]

ldr r1,=AGB_IRQVECT
adr r2,irqhandler
str r2,[r1]

mov r0,#1
str r0,[r5,#REG_IME]

@cls:

@argument_menu
@r6 = argument_count (multiplied by 12)
@r7 = arguments_array
@during: r0 = selected (multiplied by 12), r1 = keyinput, r2 = old_keyinput, r3 = old_select, r4 = REG_BASE + REG_P1, r8 = dpad, r9 = lastpad, r10 = repeatcount
menu:
mov r0,#0
@if magic key, old menu position should be saved, so reload it
ldr r2,=MAGIC_KEY
ldr r4,=MAGIC_KEY_OFFSET
ldr r1,[r4],#4
cmp r1,r2
ldreq r0,[r4],#4
ldreq r3,[r4],#4
eor r3,r3,r2
cmp r3,r0
movne r0,#0
bl waitframe
bl drawmenu

ldr r4,=REG_BASE
add r4,r4,#REG_P1
ldr r2,[r4]
mvn r2,r2

mov r3,#0
mov r9,#0
mov r10,#0

mov r11,#0

menu_loop:
bl waitframe

add r11,r11,#1
and r11,r11,#63
subs r12,r11,#31
rsblt r12,r12,#0
lsr r12,r12,#1
rsb r5,r12,#16
add r12,r12,r5,lsl#8
mov r5,#REG_BASE
strh r12,[r5,#REG_BLDALPHA]

bl getinput

tst r1,#RIGHT
beq not_right
add r0,r0,#(10*12)
cmp r0,r6
blt not_right
sub r0,r6,#(12)

not_right:
tst r1,#LEFT
beq not_left
subs r0,r0,#(10*12)
bgt not_left
mov r0,#0

not_left:
tst r1,#UP
beq not_up
subs r0,r0,#(12)
bgt not_up
mov r0,#0

not_up:
tst r1,#DOWN
beq not_down
add r0,r0,#(12)
cmp r0,r6
blt not_down
sub r0,r6,#(12)

not_down:
cmp r0,r3
beq no_drawmenu
bl drawmenu
mov r3,r0

no_drawmenu:

tst r1,#(A_BTN+B_BTN+START)
beq menu_loop

tst r1,#(B_BTN)
beq not_b
adr r1,reset_code
ldr r1,[r1]
cmp r1,#0
beq not_b

@b pressed and reset code found, so jump to reset code
bx r1

not_b:

ldr r2,=0x03FF

wait_for_no_input:
bl waitframe
bl getinput
tst r1,r2
bne wait_for_no_input

ldr r2,=MAGIC_KEY
ldr r4,=MAGIC_KEY_OFFSET
str r2,[r4],#4
str r0,[r4],#4
eor r3,r0,r2
str r3,[r4],#4

load_details:
add r7,r7,r0
ldr r0,[r7],#4 
ldr r1,[r7],#4
ldr r2,[r7]

copy_arguments:
@r0 = file_size, r1 = pointer_to_file, r2 = pointer_to_filename
ldr r5,=ARGUMENTS_OFFSET
str r0,[r5],#4
str r1,[r5],#4
mov r3,#0xBE
strb r3,[r5],#1
mov r3,#0xBA
strb r3,[r5],#1
mov r3,#0xB0
strb r3,[r5],#1
mov r3,#0xFA
strb r3,[r5],#1
mov r0,#2
str r0,[r5],#4
adr r4,cmdline
ldr r4,[r4]
bl strcpy
mov r3,#0x2f @ '/'
strb r3,[r5],#1
mov r4,r2
bl strcpy

@return to plugin
return_to_plugin:
mov r0,#0
str r0,[r5,#REG_IME]
mov r0,#0x0000
strh r0,[r5,#REG_DISPSTAT]
@ldr r0,=0xFFFF0000
mov r0,#0
str r0,[r5,#REG_IE]

@clear iwram
mov r0,#0x0
ldr r1,=0x03000000
ldr r2,=(1<<24|0x2000)
swi 0x0C0000 @CpuFastSet

@fix throbber
ldr r0,=((16<<0)|(0<<8))
mov r5,#REG_BASE
strh r0,[r5,#REG_BLDALPHA]

@only do a register ram reset on non-multiboot
adr r0,return_point
ldr r0,[r0]
tst r0,#0x02000000
bne 0f

mov r0,#0xfc
swi 0x010000

adr r0,return_point
ldr r0,[r0]

0:
bx r0

variables:

reset_code:					.word 0x00000000
return_point:				.word 0x00000000
plugin_cpufastset:	.word 0x00000000
font:								.word 0x00000000
fontpal:						.word 0x00000000
background_char:		.word 0x00000000
background_pal:			.word 0x00000000
background_map:			.word 0x00000000
cmdline:						.word 0x00000000
argument_count:			.word 0x00000000
arguments_array:		.word 0x00000000

@array_entry:
@.word file_size
@.word file_pointer
@.word zero_terminated_filename_pointer

strcpy:
ldrb r1,[r4],#1
strb r1,[r5],#1
cmp r1,#0
bne strcpy
bx lr

drawmenu:
@r0 = selected, r6 = argument_count, r7 = arguments_array
stmfd sp!,{r0-r5,lr}

bl drawmenu_compute_offset

0:
mov r3,#0
cmp r2,r0
moveq r3,#1
add r4,r7,r2
ldr r4,[r4,#8]
bl drawtext
add r2,r2,#12
add r1,r1,#1
cmp r1,#21
bge 2f
cmp r2,r6
bge 1f
b 0b

1:

mov r3,#-1
bl drawtext
add r1,r1,#1
cmp r1,#21
blt 1b

2:

ldmfd sp!,{r0-r5,lr}
bx lr

drawmenu_compute_offset:
@enter: r0 = selected * 12, r6 = argument_count * 12
@return: r1 = first draw line, r2 = first selected line
stmfd sp!,{r0,r3-r5,lr}

cmp r6,#(20*12)
bge 0f

mov r2,#4
mov r5,#REG_BASE
strh r2,[r5,#REG_BG0VOFS]
strh r2,[r5,#REG_BG1VOFS]

mov r0,r6
mov r1,#(12*2)
swi 0x060000 @ DIV

rsb r1,r0,#10
mov r2,#0
b 1f

0:
@mov r4,r0

@lsl r0,r0,#3
@mov r1,r6
@swi 0x060000 @ DIV
@mov r5,#REG_BASE
@strh r0,[r5,#REG_BG0VOFS]
@strh r0,[r5,#REG_BG1VOFS]

@mov r1,#21
@mul r0,r1,r4
@mov r1,r6
@swi 0x060000 @ DIV
@mov r1,#12
@mul r0,r1,r0

@mov r1,#0
@sub r2,r4,r0

adr r1,argument_count
ldr r2,[r1]

sub r1,r2,#(20)
mul r0,r1,r0
lsl r0,r0,#3
sub r1,r2,#(1)
mov r2,#(12)
mul r1,r2,r1
swi 0x060000 @ DIV

lsr r2,r0,#3
and r0,r0,#7
mov r5,#REG_BASE
strh r0,[r5,#REG_BG0VOFS]
strh r0,[r5,#REG_BG1VOFS]

mov r1,#12
mul r2,r1,r2
mov r1,#0
@mov r2,r4

1:
ldmfd sp!,{r0,r3-r5,lr}
bx lr


drawtext:
@r1 = row, r3 = hilight, r4 = string
stmfd sp!,{r0-r5,lr}
ldr r5,=SCREENBASE
ldr r2,=THROBBERBASE
add r5,r5,r1,LSL#6
add r5,r5,#2
add r2,r2,r1,LSL#6
cmp r3,#0
mov r3,#31
blt 1f
mov r1,#0xe000
moveq r0,#0xe000
movne r0,#0xf000
addne r1,r1,#0x2a
strh r1,[r2],#2

0:
ldrb r1,[r4],#1
cmp r1,#0
beq 2f
orr r2,r0,r1
strh r2,[r5],#2
sub r3,r3,#1
cmp r3,#0
bne 0b
b 3f
1:
mov r0,#0xe000
strh r0,[r5],#2
subs r3,r3,#1
2:
cmp r3,#0
bne 1b
3:

ldmfd sp!,{r0-r5,lr}
bx lr

getinput:
@entry: r1 = keyinput, r2 = old_keyinput, r3 = old_select, r4 = REG_BASE + REG_P1, r8 = dpad, r9 = lastpad, r10 = repeatcount
ldr r5,[r4]
mvn r5,r5
eor r1,r2,r5
and r1,r1,r5
mov r2,r5
and r8,r5,#(LEFT+RIGHT+UP+DOWN)
cmp r8,r9
bne 0f
add r10,r10,#1
cmp r10,#25
movlt r8,#0
tst r10,#3
moveq r8,#0
b 1f
0:
mov r10,#0
mov r9,r8
1:
and r1,r1,#(A_BTN+B_BTN+START)
orr r1,r1,r8 
bx lr

waitframe:
stmfd sp!,{r0-r3}
mov r0,#0
mov r1,#1
swi 0x040000
ldmfd sp!,{r0-r3}
bx lr

irqhandler:
mov r2,#REG_BASE
mov r3,#REG_BASE
ldr r1,[r2,#REG_IE]!
and r1,r1,r1,lsr#16 @ r1=IE&IF
ldrh r0,[r3,#-8]
orr r0,r0,r1
strh r0,[r3,#-8]
strh r1,[r2,#2]
bx lr

@APlib ARM7 decompressor by Dan Weiss, based on the original C version
@Takes in raw apacked data, NOT data created by the 'safe' compressor.

src				.req r0
dest			.req r1
byte			.req r2
mask			.req r3
gamma			.req r4
lwm				.req r6
recentoff	.req r7
temp			.req r8

@r0 = src
@r1 = dest
@r2 = byte
@r3 = rotating bit mask
@r4 = increasing gamma
@r6 = lwm
@r7 = recentoff
@r8 = lr copy/scratch

.macro GETBIT @3 instructions
movs mask,mask,ror #1
ldrcsb byte,[src],#1
tst byte,mask
.endm

.macro GETBITGAMMA @5 instructions
mov gamma,gamma,lsl #1
GETBIT
addne gamma,gamma,#1
.endm

depack:
	stmfd sp!,{r4-r10,lr}
	ldrb temp,[src],#1
	strb temp,[dest],#1
	ldr mask,=0x01010101
	b aploop_nolwm

	@depack enters here
aploop_nolwm:
	mov lwm,#0
aploop:
	GETBIT
	bne apbranch1
	ldrb temp,[src],#1
	strb temp,[dest],#1
	b aploop_nolwm
apbranch1:
	GETBIT
	beq apbranch2
	GETBIT
	beq apbranch3
	@get an offset
	mov gamma,#0
	GETBIT
	addne gamma,gamma,#1
	GETBITGAMMA
	GETBITGAMMA
	GETBITGAMMA
	cmp gamma,#0
	ldrneb gamma,[dest,-gamma]
	strb gamma,[dest],#1
	b aploop_nolwm
apbranch3:
	@use 7 bit offset, length = 2 or 3
	@if a zero is encountered here, it's EOF
	ldrb gamma,[src],#1
	movs recentoff,gamma,lsr #1
	beq done
	ldrcsb temp,[dest,-recentoff]
	strcsb temp,[dest],#1
	ldrb temp,[dest,-recentoff]
	strb temp,[dest],#1
	ldrb temp,[dest,-recentoff]
	strb temp,[dest],#1
	mov lwm,#1
	b aploop
apbranch2:
	@use a gamma code * 256 for offset, another gamma code for length

	bl ap_getgamma
	sub gamma,gamma,#2
	cmp lwm,#0
	bne ap_is_lwm
	mov lwm,#1
	cmp gamma,#0
	bne ap_not_zero_gamma

	@if gamma code is 2, use old recent offset, and a new gamma code for length
	bl ap_getgamma
copyloop1:
	ldrb temp,[dest,-recentoff]
	strb temp,[dest],#1
	subs gamma,gamma,#1
	bne copyloop1
	b aploop
	
ap_not_zero_gamma:
	sub gamma,gamma,#1
ap_is_lwm:
	ldrb temp,[src],#1
	add recentoff,temp,gamma,lsl #8
	bl ap_getgamma
	@gamma=length
	cmp recentoff,#32000
	addge gamma,gamma,#1
	cmp recentoff,#1280
	addge gamma,gamma,#1
	cmp recentoff,#128
	addlt gamma,gamma,#2
copyloop2:
	ldrb temp,[dest,-recentoff]
	strb temp,[dest],#1
	subs gamma,gamma,#1
	bne copyloop2
	b aploop

ap_getgamma:
	mov gamma,#1
ap_getgammaloop:
	GETBITGAMMA
	GETBIT
	bne ap_getgammaloop
	bx lr

done:
	ldmfd sp!,{r4-r10,lr}
	bx lr

